import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.distributions import Categorical
import gymnasium as gym

from utils import episode_reward_plot
from vpg import compute_returns, ActorNetwork


def compute_advantages(returns, values):
    """ Compute episode advantages based on precomputed episode returns.

    Parameters
    ----------
    returns : list of float
        Episode returns calculated with compute_returns.
    values: list of float
        Critic outputs for the states visited during the episode

    Returns
    -------
    list of float
        Episode advantages.
    """

    # TODO (4.)
    return None


def compute_generalized_advantages(rewards, values, next_value, discount, lamb):
    """ Compute generalized advantages (GAE) of the episode.

    Parameters
    ----------
    rewards : list of float
        Episode rewards.
    values: list of float
        Episode state values.
    next_value : float
        Value of state the episode ended in. Should be 0.0 for terminal state, critic output otherwise.
    discount : float
        Discount factor.
    lamb: float
        Lambda parameter of GAE.

    Returns
    -------
    list of float
        Generalized advanteges of the episode.
    """

    # TODO (8.)
    return None


class TransitionMemory:
    """Datastructure to store episode transitions and perform return/advantage/generalized advantage calculations (GAE) at the end of an episode."""

    def __init__(self, gamma, lamb, use_gae):

        # TODO Copy from vpg.py and extend
        pass

    def put(self, obs, action, reward, logprob, value):
        """Put a transition into the memory."""

        # TODO Copy from vpg.py and extend
        pass

    def get(self):
        """Get all stored transition attributes in the form of lists."""

        # TODO Copy from vpg.py and extend
        return None

    def clear(self):
        """Reset the transition memory."""

        # TODO Copy from vpg.py and extend
        pass

    def finish_trajectory(self, next_value):
        """Call on end of an episode. Will perform episode return and advantage or generalized advantage estimation.

        Parameters
        ----------
        next_value:
            The value of the state the episode ended in. Should be 0.0 for terminal state, critic output otherwise.
        """

        # TODO Copy from vpg.py and extend
        pass


class CriticNetwork(nn.Module):
    """Neural Network used to learn the state-value function."""

    def __init__(self, num_observations):
        super(CriticNetwork, self).__init__()

        # TODO (1.)
        pass

    def forward(self, obs):

        # TODO
        return None


class ActorCritic:
    """The Actor-Critic approach."""

    # Advantage: bs 200, lr 0.005
    # GAE: bs 500, critic lr 0.001, lr 0.01
    def __init__(self, env, batch_size=500, gamma=0.99, lamb=0.99, lr=0.01, use_gae=True):
        """ Constructor.

        Parameters
        ----------
        env : gym.Environment
            The object of the gym environment the agent should be trained in.
        batch_size : int, optional
            Number of transitions to use for one opimization step.
        gamma : float, optional
            Discount factor.
        lamb : float, optional
            Lambda parameters of GAE.
        lr : float, optional
            Learning rate used for actor and critic Adam optimizer.
        """

        if isinstance(env.action_space, gym.spaces.Box):
            raise NotImplementedError('Continous actions not implemented!')

        self.obs_dim, self.act_dim = env.observation_space.shape[0], env.action_space.n
        self.batch_size = batch_size
        self.env = env
        self.memory = TransitionMemory(gamma, lamb, use_gae)

        self.actor_net = ActorNetwork(self.obs_dim, self.act_dim)
        self.critic_net = CriticNetwork(self.obs_dim)
        self.optim_actor = optim.Adam(self.actor_net.parameters(), lr=lr)
        self.optim_critic = optim.Adam(self.critic_net.parameters(), lr=lr)

    def learn(self, total_timesteps):
        """Train the actor-critic.

        Parameters
        ----------
        total_timesteps : int
            Number of timesteps to train the agent for.
        """
        obs, _ = self.env.reset()

        # For plotting
        overall_rewards = []
        episode_rewards = []

        episode_counter = 0
        for timestep in range(1, total_timesteps + 1):

            # TODO (7.a)
            terminated, truncated = None, None

            if terminated or truncated:

                # TODO (7.b)
                pass

            if (timestep - episode_counter) == self.batch_size:

                # TODO (7.c)

                # Clear memory
                self.memory.clear()
                episode_counter = timestep

            # Episode reward plot
            if timestep % 500 == 0:
                episode_reward_plot(overall_rewards, timestep, window_size=5, step_size=1)

    @staticmethod
    def calc_critic_loss(value_lst, return_lst):
        """Calculate critic loss for one batch of transitions."""

        # TODO (5.)
        return None

    @staticmethod
    def calc_actor_loss(logprob_lst, adv_lst):
        """Calculate actor "loss" for one batch of transitions."""

        # TODO (6.)
        return None

    def predict(self, obs, train_returns=False):
        """Sample the agents action based on a given observation.

        Parameters
        ----------
        obs : numpy.array
            Observation returned by gym environment
        train_returns : bool, optional
            Set to True to get log probability of decided action and predicted value of obs.
        """

        # TODO copy from vpg.py and adjust (3.)
        pass

        if train_returns:
            return None
        else:
            return None


if __name__ == '__main__':
    env_id = "CartPole-v1"
    use_gae = True

    _env = gym.make(env_id)
    AC = ActorCritic(_env, use_gae=use_gae)
    AC.learn(100000)
