import torch
import torch.optim as optim
import torch.nn.functional as F
from torch.distributions import Categorical
import gymnasium as gym

from utils import episode_reward_plot
from vpg import compute_returns, ActorNetwork
from a2c import CriticNetwork, compute_advantages, compute_generalized_advantages


class TransitionMemory:
    """Datastructure to store episode transitions and perform return/advantage/generalized advantage calculations (GAE) at the end of an episode."""

    def __init__(self, gamma, lamb, use_gae):

        # TODO Copy from a2c.py and extend (1.)
        pass

    def put(self, obs, action, reward, logprob, logprob_old, value):
        """Put a transition into the memory."""

        # TODO Copy from a2c.py and extend
        pass

    def get(self):
        """Get all stored transition attributes in the form of lists."""

        # TODO Copy from a2c.py and extend
        return None

    def clear(self):
        """Reset the transition memory."""

        # TODO Copy from a2c.py and extend
        pass

    def finish_trajectory(self, next_value):
        """Call on end of an episode. Will perform episode return and advantage or generalized advantage estimation.

        Parameters
        ----------
        next_value:
            The value of the state the episode ended in. Should be 0.0 for terminal state, critic output otherwise.
        """

        # TODO Copy from a2c.py and extend
        pass


class ProximalPolicyOptimization:
    """The Actor-Critic approach."""

    def __init__(self, env, batch_size=500, gamma=0.99, lamb=0.99, lr=0.01, eps=0.5, use_gae=True):
        """ Constructor.

        Parameters
        ----------
        env : gym.Environment
            The object of the gym environment the agent should be trained in.
        batch_size : int, optional
            Number of transitions to use for one opimization step.
        gamma : float, optional
            Discount factor.
        lamb : float, optional
            Lambda parameters of GAE.
        lr : float, optional
            Learning rate used for actor and critic Adam optimizer.
        """

        if isinstance(env.action_space, gym.spaces.Box):
            raise NotImplementedError('Continous actions not implemented!')

        self.obs_dim, self.act_dim = env.observation_space.shape[0], env.action_space.n
        self.batch_size = batch_size
        self.env = env
        self.eps = eps
        self.memory = TransitionMemory(gamma, lamb, use_gae)

        self.actor_net = ActorNetwork(self.obs_dim, self.act_dim)
        self.critic_net = CriticNetwork(self.obs_dim)
        self.optim_actor = optim.Adam(self.actor_net.parameters(), lr=lr)
        self.optim_critic = optim.Adam(self.critic_net.parameters(), lr=lr)

    def learn(self, total_timesteps):
        """Train the actor-critic.

        Parameters
        ----------
        total_timesteps : int
            Number of timesteps to train the agent for.
        """
        obs, _ = self.env.reset()

        # For plotting
        overall_rewards = []
        episode_rewards = []

        logprob_old = None

        episode_counter = 0
        for timestep in range(1, total_timesteps + 1):

            # TODO Copy from a2c.py and make small adjustments (1.)

            terminated, truncated = None, None

            if terminated or truncated:

                pass

            if (timestep - episode_counter) == self.batch_size:

                pass

                # Clear memory
                self.memory.clear()
                episode_counter = timestep

            # Episode reward plot
            if timestep % 500 == 0:
                episode_reward_plot(overall_rewards, timestep, window_size=5, step_size=1)

    @staticmethod
    def calc_critic_loss(value_lst, return_lst):
        """Calculate critic loss for one batch of transitions."""

        # TODO Copy from a2c.py
        return None

    @staticmethod
    def calc_actor_loss(logprob_lst, logprob_old_lst, adv_lst, eps):
        """Calculate actor "loss" for one batch of transitions."""

        # TODO (2.)
        return None

    def predict(self, obs, train_returns=False):
        """Sample the agents action based on a given observation.

        Parameters
        ----------
        obs : numpy.array
            Observation returned by gym environment
        train_returns : bool, optional
            Set to True to get log probability of decided action and predicted value of obs.
        """

        # TODO copy from a2c.py

        if train_returns:
            return None
        else:
            return None


if __name__ == '__main__':
    env_id = "CartPole-v1"
    use_gae = True

    _env = gym.make(env_id)
    PPO = ProximalPolicyOptimization(_env, use_gae=use_gae)
    PPO.learn(100000)
